#-*- coding:utf-8 -*-
#
# database/__init__.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Module dedicated to logging the simulations and networks generated via the
library.

Depending on the settings in `$HOME/.nngt.conf`, the data will either be stored
in a in a SQL database or in CSV files.

Content
=======
"""

import os, errno

from peewee import *
from playhouse.db_url import connect

import nngt


__all__ = [
	'get_results',
	'is_clear',
	'log_simulation_end',
	'log_simulation_start',
	'reset',
]



# --------------------------------------- #
# Parse config file and generate database #
# --------------------------------------- #

def _set_main_db():
    if nngt.get_config("db_url") is None or nngt.get_config("db_to_file"):
        # check for db_folder
        abs_dbfolder = os.path.abspath(
            os.path.expanduser(nngt.get_config("db_folder")))
        if not os.path.isdir(abs_dbfolder):
            try:
                os.mkdir(abs_dbfolder)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        # create database
        db_file = abs_dbfolder + "/" + nngt.get_config("db_name") + ".db"

        nngt._main_db = SqliteDatabase(
            db_file,
            pragmas=(('journal_mode', 'wal'), ('cache_size', -1024 * 64),
                     ('foreign_keys', 'on'),))
    else:
        nngt._main_db = connect(
            nngt.get_config('db_url'), fields={'longblob': 'longblob'})


# ----------------- #
# Set NNGT database #
# ----------------- #

_set_main_db()

# IMPORTANT: this MUST come after the call to _set_main_db
from .db_main import NNGTdb as _db

if nngt._db is None:
    nngt._db = _db()

# --------- #
# Functions #
# --------- #

get_results          = nngt._db.get_results
is_clear             = nngt._db.is_clear
log_simulation_end   = nngt._db.log_simulation_end
log_simulation_start = nngt._db.log_simulation_start
reset                = nngt._db.reset



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from itertools import chain

from xml.dom.minidom import parse

from svg.path import (parse_path, CubicBezier, QuadraticBezier, Arc,
                      Move, Close, Path)

import shapely
from shapely.affinity import scale, affine_transform, translate
from shapely.geometry import Point, Polygon

import numpy as np

from .shape import Shape


'''
Shape generation from SVG files.
'''


__all__ = ["polygons_from_svg"]


# predefined svg shapes and their parameters

_predefined = {
    'path': None,
    'ellipse': ("cx", "cy", "rx", "ry"),
    'circle': ("cx", "cy", "r"),
    'rect': ("x", "y", "width", "height")
}

_valid_nodes = _predefined.keys()


def polygons_from_svg(filename, interpolate_curve=50, parent=None,
                      return_points=False):
    '''
    Generate :class:`shapely.geometry.Polygon` objects from an SVG file.
    '''
    svg = parse(filename)
    elt_structs = {k: [] for k in _valid_nodes}
    elt_points = {k: [] for k in _valid_nodes}

    # get the properties of all predefined elements
    for elt_type, elt_prop in _predefined.items():
        _build_struct(svg, elt_structs[elt_type], elt_type, elt_prop)

    # build all shapes
    polygons = []
    for elt_type, instructions in elt_structs.items():
        for struct in instructions:
            polygon, points = _make_polygon(
                elt_type, struct, parent=parent, return_points=True)
            polygons.append(polygon)
            elt_points[elt_type].append(points)

    if return_points:
        return polygons, elt_points

    return polygons


# ----- #
# Tools #
# ----- #


def _get_closed_subpaths(path):
    '''
    Generates all closed subpaths raises error if open subpaths exist.

    Credit to @tatarize:
    https://github.com/regebro/svg.path/issues/54#issuecomment-570101018
    '''
    segments = None
    for p in path:
        if isinstance(p, Move):
            if segments is not None:
                raise RuntimeError("Only closed shapes accepted.")
            segments = []

        segments.append(p)

        if isinstance(p, Close):
            yield Path(*segments)
            segments = None


def _get_points(path, interpolate_curve=50):
    ''' Get points from path. '''
    points = []

    for item in path:
        if isinstance(item, (Arc, CubicBezier, QuadraticBezier)):
            istart = 1. / interpolate_curve
            for frac in np.linspace(istart, 1, interpolate_curve):
                points.append(
                    (item.point(frac).real, item.point(frac).imag))
        else:
            points.append((item.start.real, item.start.imag))

    return points


def _get_outer_shell(paths_points):
    ''' Returns the index of the container subpath '''
    minx, maxx, miny, maxy = np.inf, -np.inf, np.inf, -np.inf

    container_idx = None

    for i, pp in enumerate(paths_points):
        arr = np.array(pp)

        x_min = np.min(arr[:, 0])
        x_max = np.max(arr[:, 0])
        y_min = np.min(arr[:, 1])
        y_max = np.max(arr[:, 1])

        winner = True

        if x_min <= minx:
            minx = x_min
        else:
            winner = False

        if x_max >= maxx:
            maxx = x_max
        else:
            winner = False

        if y_min <= miny:
            miny = y_min
        else:
            winner = False

        if y_max >= maxy:
            maxy = y_max
        else:
            winner = False

        if winner:
            container_idx = i

    return container_idx


def _build_struct(svg, container, elt_type, elt_properties):
    root    = svg.documentElement

    for elt in root.getElementsByTagName(elt_type):
        struct = {
            "transf": [],
            "transfdata": []
        }

        parent = elt.parentNode

        while parent is not None:
            _get_transform(parent, struct)
            parent = parent.parentNode

        _get_transform(elt, struct)
        
        if elt_type == 'path':
            path, trans = elt.getAttribute('d'), None
            struct["path"] = path
        else:
            for item in elt_properties:
                struct[item] = float(elt.getAttribute(item))

        container.append(struct)


def _make_polygon(elt_type, instructions, parent=None, interpolate_curve=50,
                  return_points=False):
    container = None
    shell     = []  # outer points defining the polygon's outer shell
    holes     = []  # inner points defining holes
    idx_start = 0

    if elt_type == "path":  # build polygons from custom paths
        path_data = parse_path(instructions["path"])
        subpaths  = [subpath for subpath in
                     _get_closed_subpaths(path_data)]
        points    = [_get_points(subpath) for subpath in subpaths]

        # get the container
        idx_container = _get_outer_shell(points)
        shell         = np.array(points[idx_container])

        # get the holes and make the shape
        holes     = [pp for i, pp in enumerate(points) if i!= idx_container]
        container = Polygon(shell, holes=holes)
    elif elt_type == "ellipse":  # build ellipses
        circle = Point((instructions["cx"], instructions["cy"])).buffer(1)
        rx, ry = instructions["rx"], instructions["ry"]
        container = scale(circle, rx, ry)
    elif elt_type == "circle":   # build circles
        r = instructions["r"]
        container = Point((instructions["cx"], instructions["cy"])).buffer(r)
    elif elt_type == "rect":     # build rectangles
        x, y = instructions["x"], instructions["y"]
        w, h = instructions["width"], instructions["height"]
        shell = np.array([(x, y), (x + w, y), (x + w, y + h), (x, y + h)])
        container = Polygon(shell)
    else:
        raise RuntimeError("Unexpected element type: '{}'.".format(elt_type))

    # transforms
    nn, dd = instructions["transf"][::-1], instructions["transfdata"][::-1]
    for name, data in zip(nn, dd):
        if name == "matrix":
            container = affine_transform(container, data)
        elif name == "translate":
            container = translate(container, *data)

    # y axis is inverted in SVG, so make mirror transform
    container = affine_transform(container, (1, 0, 0, -1, 0, 0))
    shell     = np.array(container.exterior.coords)

    if return_points:
        return container, shell

    return container


def _get_transform(obj, tdict):
    ''' Get the transformation properties and name into `tdict` '''
    try:
        if obj.hasAttribute("transform"):
            trans = obj.getAttribute('transform')
            if trans.startswith("translate"):
                start = trans.find("(") + 1
                stop  = trans.find(")")
                tdict["transf"].append("translate")
                tdict["transfdata"].append(
                    [float(f) for f in trans[start:stop].split(",")])
            elif trans.startswith("matrix"):
                start = trans.find("(") + 1
                stop  = trans.find(")")
                trans = [float(f)
                         for f in trans[start:stop].split(",")]
                tdict["transf"].append("matrix")
                tdict["transfdata"].append(trans)
            else:
                raise RuntimeError("Uknown transform: " + trans)
    except:
        pass



#-*- coding:utf-8 -*-
#
# lib/logger.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Logging for the NNGT module """

import inspect
import logging
import os
import warnings

from datetime import date

import scipy.sparse as ssp

import nngt
from .test_functions import mpi_checker


# ignore scipy.sparse efficiency warnings
warnings.filterwarnings("ignore", category=ssp.SparseEfficiencyWarning)

# ignore igraph unconnected network warnings
warnings.filterwarnings("ignore", message="Couldn't reach some vertices at")

# check that log folder exists, otherwise create it

nngt._config["log_folder"] = os.path.expanduser(nngt._config["log_folder"])
if not os.path.isdir(nngt._config["log_folder"]):
    os.mkdir(nngt._config["log_folder"])


# configure logger

def _init_logger(logger):
    logger.handlers = []  # necessary for ipython
    logger.setLevel(logging.INFO)
    logConsoleFormatter = logging.Formatter(
        '[%(levelname)s @ %(name)s]: %(message)s')
    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logConsoleFormatter)
    consoleHandler.setLevel(nngt._config["log_level"])
    logger.addHandler(consoleHandler)


def _configure_logger(logger):
    logger.setLevel(logging.INFO)
    is_writing_to_file = False
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.setLevel(nngt._config["log_level"])
        elif isinstance(handler, logging.FileHandler):
            is_writing_to_file = True
    if nngt._config["log_to_file"]:
        _log_to_file(logger, not is_writing_to_file)


# initialize file log

def _log_to_file(logger, create_writer):
    if create_writer:
        logFileFormatter = logging.Formatter(
            '[%(levelname)s @ %(funcName)s] %(asctime)s:\n\t%(message)s')
        today = date.today()
        fileName = "/nngt_{}-{}-{}".format(today.month, today.day, today.year)
        fileHandler = logging.FileHandler(
            "{}/{}.log".format(nngt._config["log_folder"], fileName))
        fileHandler.setFormatter(logFileFormatter)
        fileHandler.setLevel(logging.DEBUG)
        logger.addHandler(fileHandler)


# log message

@mpi_checker(logging=True)
def _log_message(logger, level, message):

    stack = inspect.stack()

    fn = stack[-1][1]
    ln = stack[-1][2]

    location = 'from ' + fn[fn.rfind("/") + 1:] + ' (L{}) - '.format(ln)

    message = location + message

    if level == 'DEBUG':
        logger.debug(message)
    elif level == 'WARNING':
        logger.warning(message)
    elif level == 'INFO':
        logger.info(message)
    else:
        logger.critical(message)



#-*- coding:utf-8 -*-
#
# plot/__init__.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Functions for plotting graphs and graph properties.

The following features are provided:

* basic graph plotting
* plotting the distribution of some attribute over the graph
* animation of some recorded activity
"""

import warnings as _warn
_warn.filterwarnings("ignore", module="matplotlib")

import matplotlib as _mpl

import nngt as _nngt

# module import

from .custom_plt import palette_continuous, palette_discrete, markers
from .animations import Animation2d, AnimationNetwork
from .plt_networks import chord_diagram, draw_network, hive_plot, library_draw
from .plt_properties import *
from . import plt_properties as _plt_prop


__all__ = [
    "Animation2d",
    "AnimationNetwork",
    "chord_diagram",
    "draw_network",
    "hive_plot",
    "library_draw",
    "palette_continuous",
    "palette_discrete",
    "markers",
]


if _nngt._config["mpl_backend"] is not None:
    _mpl.use(_nngt._config["mpl_backend"])


__all__.extend(_plt_prop.__all__)



#-*- coding:utf-8 -*-

# test_graphclasses.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the main methods of the :class:`~nngt.Graph` class and its subclasses.
"""

import unittest
import pytest

import numpy as np
import scipy.sparse as ssp

import nngt
import nngt.generation as ng

from base_test import TestBasis, XmlHandler, network_dir
from tools_testing import foreach_graph


# ---------- #
# Test class #
# ---------- #

class TestGraphClasses(TestBasis):

    '''
    Class testing the main methods of :class:`~nngt.Graph` and its subclasses.
    '''

    matrices = {}
    mat_gen = {
        "from_scipy_sparse_rand": ssp.rand,
        "from_numpy_randint": np.random.randint
    }

    @property
    def test_name(self):
        return "test_graphclasses"

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def gen_graph(self, graph_name):
        di_instructions = self.parser.get_graph_options(graph_name)
        mat = self.mat_gen[graph_name](**di_instructions)
        self.matrices[graph_name] = mat
        graph = nngt.Graph.from_matrix(mat)
        graph.set_name(graph_name)
        return graph, di_instructions

    @foreach_graph
    def test_adj_mat(self, graph, **kwargs):
        '''
        When generating graphs from :class:`numpy.ndarray`s or
        :mod:`scipy.sparse` matrices, check that the result of
        graph.adjacency_matrix() is the same as the initial matrix.
        '''
        ref_result = ssp.csr_matrix(self.matrices[graph.name])
        computed_result = graph.adjacency_matrix(weights=True)
        self.assertTrue(
            (ref_result != computed_result).nnz == 0,
            "AdjMat test failed for graph {}:\nref = {} vs exp {}\
            ".format(graph.name, ref_result, computed_result))

    @foreach_graph
    def test_copy_clear(self, graph, **kwargs):
        '''
        Test that the copied graph is indeed the same as the original, but that
        all its properties are deep copies.
        Then check that clear_edges() removes all edges and no nodes.
        '''
        ref_result = (graph.node_nb(), graph.edge_nb(), graph.node_nb(), 0)
        copied = graph.copy()
        self.assertIsNot(copied, graph)
        computed_result = [copied.node_nb(), copied.edge_nb()]
        copied.clear_all_edges()
        computed_result.extend((copied.node_nb(), copied.edge_nb()))
        self.assertEqual(
            ref_result, tuple(computed_result),
            "Copy test failed for graph {}:\nref = {} vs exp {}\
            ".format(graph.name, ref_result, computed_result))


@pytest.mark.mpi_skip
def test_structure_graph():
    gclass = (nngt.Group, nngt.NeuralGroup)
    sclass = (nngt.Structure, nngt.NeuralPop)
    nclass = (nngt.Graph, nngt.Network)

    for gc, sc, nc in zip(gclass, sclass, nclass):
        room1 = gc(25, neuron_type=1)
        room2 = gc(50, neuron_type=1)
        room3 = gc(40, neuron_type=1)
        room4 = gc(35, neuron_type=1)

        names = ["R1", "R2", "R3", "R4"]

        kwargs = {"with_models": False} if sc == nngt.NeuralPop else {}
        struct = sc.from_groups((room1, room2, room3, room4), names, **kwargs)

        kwargs = ({"population": struct} if nc == nngt.Network
                  else {"structure": struct})

        g = nc(**kwargs)

        # connect groups
        for room in struct:
            ng.connect_groups(g, room, room, "all_to_all")

        d1 = 5
        ng.connect_groups(g, room1, room2, "erdos_renyi", avg_deg=d1)
        ng.connect_groups(g, room1, room3, "erdos_renyi", avg_deg=d1)
        ng.connect_groups(g, room1, room4, "erdos_renyi", avg_deg=d1)

        d2 = 5
        ng.connect_groups(g, room2, room3, "erdos_renyi", avg_deg=d2)
        ng.connect_groups(g, room2, room4, "erdos_renyi", avg_deg=d2,
                                       weights=2)

        d3 = 20
        ng.connect_groups(g, room3, room1, "erdos_renyi", avg_deg=d3)

        d4 = 10
        ng.connect_groups(g, room4, room3, "erdos_renyi", avg_deg=d4)

        # get structure graph
        sg = g.get_structure_graph()

        assert sg.node_nb() == len(struct)

        eset = set([tuple(e) for e in sg.edges_array])
        expected = [
            (0, 0), (0, 1), (0, 2), (0, 3),
            (1, 1), (1, 2), (1, 3),
            (2, 0), (2, 2),
            (3, 2), (3, 3)
        ]

        assert eset == set(expected)

        # check weights
        w1 = room1.size * d1
        w2 = room2.size * d2
        w3 = room3.size * d3
        w4 = room4.size * d4

        expected_weights = [
            room1.size*(room1.size - 1), w1, w1, w1,
            room2.size*(room2.size - 1), w2, w2*2,
            w3, room3.size*(room3.size - 1),
            w4, room4.size*(room4.size - 1)
        ]

        assert np.array_equal(sg.get_weights(edges=expected), expected_weights)


@pytest.mark.mpi_skip
def test_autoclass():
    '''
    Check that Graph is automatically converted to Network or SpatialGraph
    if the relevant arguments are provided.
    '''
    pop = nngt.NeuralPop.exc_and_inhib(100)

    g = nngt.Graph(population=pop)

    assert isinstance(g, nngt.Network)

    shape = nngt.geometry.Shape.disk(50.)

    g = nngt.Graph(shape=shape)

    assert isinstance(g, nngt.SpatialGraph)

    g = nngt.Graph(population=pop, shape=shape)

    assert isinstance(g, nngt.SpatialNetwork)


# ---------- #
# Test suite #
# ---------- #

if not nngt.get_config('mpi'):
    suite = unittest.TestLoader().loadTestsFromTestCase(TestGraphClasses)

    if __name__ == "__main__":
        unittest.main()
        test_structure_graph()
        test_autoclass()



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Node and edge attributes '''

import numpy as np

import nngt
import nngt.generation as ng


''' -------------- #
# Generate a graph #
# -------------- '''

num_nodes = 1000
avg_deg   = 25

graph = ng.erdos_renyi(nodes=num_nodes, avg_deg=avg_deg)


''' ----------------- #
# Add node attributes #
# ----------------- '''

# Let's make a network of animals where nodes represent either cats or dogs.
# (no discrimination against cats or dogs was intended, no animals were harmed
# while writing or running this code)
animals  = ["cat" for _ in range(600)]  # 600 cats
animals += ["dog" for _ in range(400)]  # and 400 dogs
np.random.shuffle(animals)              # which we assign randomly to the nodes

graph.new_node_attribute("animal", value_type="string", values=animals)

# Let's check the type of the first six animals
print(graph.get_node_attributes([0, 1, 2, 3, 4, 5], "animal"))

# Nodes can have attributes of multiple types, let's add a size to our animals
catsizes = np.random.normal(50, 5, 600)   # cats around 50 cm
dogsizes = np.random.normal(80, 10, 400)  # dogs around 80 cm

# We first create the attribute without values (for "double", default to NaN)
graph.new_node_attribute("size", value_type="double")

# We now have to attributes: one containing strings, the other numbers (double)
print(graph.node_attributes)

# get the cats and set their sizes
cats = graph.get_nodes(attribute="animal", value="cat")
graph.set_node_attribute("size", values=catsizes, nodes=cats)

# We set 600 values so there are 400 NaNs left
assert np.sum(np.isnan(graph.get_node_attributes(name="size"))) == 400, \
    "There were not 400 NaNs as predicted."

# None of the NaN values belongs to a cat
assert not np.any(np.isnan(graph.get_node_attributes(cats, name="size"))), \
    "Got some cats with NaN size! :'("

# get the dogs and set their sizes
dogs = graph.get_nodes(attribute="animal", value="dog")
graph.set_node_attribute("size", values=dogsizes, nodes=dogs)

# Some of the animals are part of human househols, they have therefore "owners"
# which will be represented here through a Human class.
# Animals without an owner will have an empty list as attribute.

class Human:
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return "Human<{}>".format(self.name)

# John owns all animals between 8 and 48
John    = Human("John")
animals = [i for i in range(8, 49)]

graph.new_node_attribute("owners", value_type="object", val=[])
graph.set_node_attribute("owners", val=[John], nodes=animals)

# Now suppose another human, Julie, owns all animals between 0 and 40
Julie   = Human("Julie")
animals = [i for i in range(0, 41)]

# to update the values, we need to get them to add Bob to the list
owners = graph.get_node_attributes(name="owners", nodes=animals)

for interactions in owners:
    interactions.append(Julie)

graph.set_node_attribute("owners", values=owners, nodes=animals)

# now some of the initial owners should have had their attributes updated
new_owners = graph.get_node_attributes(name="owners")
print("There are animals owned only by", new_owners[0], "others owned only by",
      new_owners[48], "and some more owned by both", new_owners[40])


''' ---------- #
# Edge weights #
# ---------- '''

# Same as for node attributes, one can give attributes to the edges
# Let's give weights to the edges depending on how often the animals interact!
# cat's interact a lot among themselves, so we'll give them high weights
cat_edges = graph.get_edges(source_node=cats, target_node=cats)

# check that these are indeed only between cats
cat_set  = set(cats)
node_set = set(np.unique(cat_edges))

assert cat_set == node_set, "Damned, something wrong happened to the cats!"

# uniform distribution of weights between 30 and 50
graph.set_weights(elist=cat_edges, distribution="uniform",
                  parameters={"lower": 30, "upper": 50})

# dogs have less occasions to interact except some which spend a lot of time
# together, so we use a lognormal distribution
dog_edges = graph.get_edges(source_node=dogs, target_node=dogs)
graph.set_weights(elist=dog_edges, distribution="lognormal",
                  parameters={"position": 2.2, "scale": 0.5})

# Cats do not like dogs, so we set their weights to -5
# Dogs like chasing cats but do not like them much either so we let the default
# value of 1
cd_edges = graph.get_edges(source_node=cats, target_node=dogs)
graph.set_weights(elist=cd_edges, distribution="constant",
                  parameters={"value": -5})

# Let's check the distribution (you should clearly see 4 separate shapes)
if nngt.get_config("with_plot"):
    nngt.plot.edge_attributes_distribution(graph, "weight")


''' ------------------- #
# Other edge attributes #
# ------------------- '''

# non-default edge attributes can be created as the node attributes
# let's create a class for humans and store it when two animals have interacted
# with the same human (the default will be an empty list if they did not)

# Alice interacted with all animals between 8 and 48
Alice   = Human("Alice")
animals = [i for i in range(8, 49)]
edges   = graph.get_edges(source_node=animals, target_node=animals)

graph.new_edge_attribute("common_interaction", value_type="object", val=[])
graph.set_edge_attribute("common_interaction", val=[Alice], edges=edges)

# Now suppose another human, Bob, interacted with all animals between 0 and 40
Bob     = Human("Bob")
animals = [i for i in range(0, 41)]
edges2  = graph.get_edges(source_node=animals, target_node=animals)

# to update the values, we need to get them to add Bob to the list
ci = graph.get_edge_attributes(name="common_interaction", edges=edges2)

for interactions in ci:
    interactions.append(Bob)

graph.set_edge_attribute("common_interaction", values=ci, edges=edges2)

# now some of the initial `edges` should have had their attributes updated
new_ci = graph.get_edge_attributes(name="common_interaction", edges=edges)
print(np.sum([0 if len(interaction) < 2 else 1 for interaction in new_ci]),
      "interactions have been updated among the", len(edges), "from Alice.")




#-*- coding:utf-8 -*-
#
# lib/test_functions.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Test functions for the NNGT """

import collections
import functools
import inspect
import warnings

from collections.abc import Container as _container
from collections.abc import Iterable as _iterable
from collections.abc import KeysView as _key_view
from collections.abc import ValuesView as _value_view

import numpy as np

import nngt

from .decorator import decorate


def deprecated(version, reason=None, alternative=None, removal=None):
    '''
    Decorator to mark deprecated functions.
    '''
    def decorator(func):
        def wrapper(func, *args, **kwargs):
            # turn off filter temporarily
            warnings.simplefilter('always', DeprecationWarning)
            message = "Function {} is deprecated since version {}"
            message = message.format(func.__name__, version)
            if reason is not None:
                message += " because " + reason + "."
            else:
                message += "."
            if removal is not None:
                message += " It will be removed in version {}.".format(removal)
            if alternative is not None:
                message += " Use " + alternative + " instead."
            warnings.warn(message, category=DeprecationWarning)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)
        return decorate(func, wrapper)
    return decorator


# ~ def constrain_argument(arg, valid_values):
    # ~ '''
    # ~ Decorator to constrain function arguments.
    # ~ '''
    # ~ def decorator(func):
        # ~ @functools.wrap(func)
        # ~ def wrapper(func, *args, **kwargs):
            # ~ arg_names = inspect.getargspec(func)[0]
            # ~ arguments = {k: v for k, v in zip(arg_names, args)}

            # ~ arguments.update(kwargs)

            # ~ if arg in arguments:
                # ~ assert arguments[arg] in valid_values, \
                    # ~ "Argument `'{}' must be among {}".format(arg, valid_values)

            # ~ return func(*args, **kwargs)
        # ~ return decorate(func, wrapper)
    # ~ return decorator


# ~ def check_arguments(arg_list):
    # ~ '''
    # ~ Decorator to constrain function arguments.
    # ~ '''
    # ~ def decorator(func):
        # ~ @functools.wrap(func)
        # ~ def wrapper(func, *args, **kwargs):
            # ~ arg_names = set(inspect.getargspec(func)[0])
            # ~ allowed   = arg_names.union(arg_list)

            # ~ for arg in kwargs:
                # ~ if arg not in allowed:
                    # ~ raise InvalidArgument(("'{}' is not a valid argument for "
                                           # ~ "function '{}'.").format(arg, func))
            # ~ return func(*args, **kwargs)
        # ~ return decorate(func, wrapper)
    # ~ return decorator


def on_master_process():
    '''
    Check whether the current code is executing on the master process (rank 0)
    if MPI is used.

    Returns
    -------
    True if rank is 0, if mpi4py is not present or if MPI is not used,
    otherwise False.
    '''
    try:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
            return True
        else:
            return False
    except ImportError:
        return True


def num_mpi_processes():
    ''' Returns the number of MPI processes (1 if MPI is not used) '''
    try:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        return comm.Get_size()
    except ImportError:
        return 1


def mpi_barrier(func=None):
    def wrapper(func, *args, **kwargs):
        try:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            comm.Barrier()
        except ImportError:
            pass

        if func is not None:
            return func(*args, **kwargs)

    # act as a real decorator
    if func is not None:
        return decorate(func, wrapper)

    # otherwise just execute the barrier
    wrapper(None)


def mpi_checker(logging=False):
    '''
    Decorator used to check for mpi and make sure only rank zero is used
    to store and generate the graph if the mpi algorithms are activated.
    '''
    def decorator(func):
        def wrapper(func, *args, **kwargs):
            # when using MPI, make sure everyone waits for the others
            try:
                from mpi4py import MPI
                comm = MPI.COMM_WORLD
                comm.Barrier()
            except ImportError:
                pass
            # check backend ("nngt" is fully parallel, not the others)
            backend = False
            if not logging:
                backend = nngt.get_config("backend") == "nngt"
            if backend or on_master_process():
                return func(*args, **kwargs)
            else:
                return None
        return decorate(func, wrapper)
    return decorator


def mpi_random(func):
    '''
    Decorator asserting that all processes start with same random seed when
    using mpi.
    '''
    def wrapper(func, *args, **kwargs):
        try:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()

            if rank == 0:
                state = np.random.get_state()
            else:
                state = None

            state = comm.bcast(state, root=0)
            np.random.set_state(state)
        except ImportError:
            pass

        return func(*args, **kwargs)

    return decorate(func, wrapper)


def nonstring_container(obj):
    '''
    Returns true for any iterable which is not a string or byte sequence.
    '''
    if isinstance(obj, (_key_view, _value_view)):
        return True

    if not isinstance(obj, _container):
        return False

    if isinstance(obj, (bytes, str)):
        return False

    return True


def is_integer(obj):
    ''' Return whether the object is an integer '''
    return isinstance(obj, (int, np.integer))


def is_iterable(obj):
    ''' Return whether the object is iterable '''
    return isinstance(obj, _iterable)


def graph_tool_check(version_min):
    '''
    Raise an error for function not working with old versions of graph-tool.
    '''
    def decorator(func):
        def wrapper(func, *args, **kwargs):
            old_graph_tool = _old_graph_tool(version_min)
            if old_graph_tool:
                raise NotImplementedError('This function is not working for '
                                          'graph-tool < ' + version_min + '.')
            else:
                return func(*args, **kwargs)
        return decorate(func, wrapper)  # to preserve the docstring info
    return decorator


def _old_graph_tool(version_min):
    '''
    Check for old versions of graph-tool for which some functions are not
    working.
    '''
    return (nngt.get_config('backend') == 'graph-tool'
            and nngt.get_config('library').__version__[:4] < version_min)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Principle
=========

Module dedicated to the description of the spatial boundaries of neuronal
cultures.
This allows for the generation of neuronal networks that are embedded in space.

The `shapely <http://toblerity.org/shapely/index.html>`_ library is used to
generate and deal with the spatial environment of the neurons.


Examples
========

Basic features
--------------

The module provides a backup ``Shape`` object, which can be used with only
the `numpy` and `scipy` libraries.
It allows for the generation of simple rectangle, disk and ellipse shapes.

.. literalinclude:: examples/backup_shape.py
   :lines: 23-

All these features are of course still available with the more advanced
``Shape`` object which inherits from :class:`shapely.geometry.Polygon`.


Complex shapes from files
-------------------------

.. literalinclude:: examples/load_culture.py
   :lines: 23-


Content
=======
"""

import logging

# shapely support

try:
    import shapely
    from shapely import speedups
    if speedups.available:
        speedups.enable()
    _shapely_support = True
    from .shape import Shape, Area
except ImportError:
    _shapely_support = False
    from .backup_shape import BackupShape as Shape

from .tools import pop_largest


__version__ = "0.8"


# -------------------- #
# Define I/O functions #
# -------------------- #

def culture_from_file(*args, **kwargs):
    raise RuntimeError("This function requires 'shapely' to work.")


def shapes_from_file(*args, **kwargs):
    raise RuntimeError("This function requires 'shapely' to work.")


if _shapely_support:
    from . import shape_io
    from .shape_io import culture_from_file as _cff
    from .shape_io import shapes_from_file as _sff
    culture_from_file = _cff
    shapes_from_file = _sff


__all__ = [
    "Area", "Shape", "culture_from_file", "pop_largest", "shapes_from_file"
]


# ------------------------------------------ #
# Try to import optional SVG and DXF support #
# ------------------------------------------ #

_logger = logging.getLogger(__name__)
from .pync_log import _log_message


try:
    from .plot import plot_shape
    __all__.append('plot_shape')
except ImportError as e:
    _log_message(_logger, "INFO", 'Could not import plotting: {}'.format(e))
